@@ -35,15 +35,63 @@ def main(ctx, config):
     default=False,
     help="Use uvicorn instead of gunicorn",
 )
+@click.option(
+    "--use-purple",
+    is_flag=True,
+    default=False,
+    help="Use purple instead of gunicorn",
+)
+@click.option(
+    "--try-fast",
+    is_flag=True,
+    default=False,
+    help="Try a faster implementation",
+)
 @click.pass_context
-def serve(ctx, use_uvicorn):
+def serve(ctx, use_uvicorn, use_purple, try_fast):
     """
     Start the http relay server
     """
-    if (
-        Config.storage.get("class", "nostr_relay.storage.db.DBStorage")
-        == "nostr_relay.storage.db.DBStorage"
-    ):
+    storage_class = Config.storage.get("class", "nostr_relay.storage.db.DBStorage")
+    if try_fast:
+        print(
+            """
+Running a demo-mode, fast server based on LMDB
+===============================================\n"""
+        )
+        storage_config = {
+            "class": "nostr_relay.storage.kv.LMDBStorage",
+            "path": "/tmp/nostr.db",
+            "map_size": 1024 * 1024 * 100,
+            "metasync": False,
+        }
+        conf_file = "/tmp/nostr-config.yaml"
+        print(
+            f"Temporarily switching config to {conf_file} and running with --use-purple"
+        )
+        Config.storage = storage_config
+        Config.purple["workers"] = 4
+        Config.purple["host"] = "127.0.0.1"
+        Config.purple["port"] = 6969
+        Config.logging["loggers"]["nostr_relay"]["level"] = "WARNING"
+        Config.config_file = conf_file
+        Config.dump(conf_file)
+        try:
+            import lmdb
+        except ImportError:
+            print("Installing lmdb...")
+            import subprocess, sys
+
+            subprocess.run(
+                [sys.executable, "-m", "pip", "install", "lmdb", "msgpack", "whoosh"]
+            )
+        from .purple import serve
+
+        print("Running on ws://127.0.0.1:6969/")
+        serve(Config)
+        return
+
+    if storage_class == "nostr_relay.storage.db.DBStorage":
         import os.path
         from alembic import config
 
@@ -54,6 +102,10 @@ def serve(ctx, use_uvicorn):
         from .web import run_with_uvicorn
 
         run_with_uvicorn()
+    elif use_purple:
+        from .purple import serve
+
+        serve(Config)
     else:
         from .web import run_with_gunicorn
 
@@ -167,12 +219,12 @@ async def update_tags(ctx, query):
     Update the tags in the tag table, from a REQ query
     """
     from .storage import get_storage
-    from .util import json
+    from .util import json_loads
 
     if not query:
         query = '[{"since": 1}]'
 
-    query = json.loads(query)
+    query = json_loads(query)
 
     async with get_storage() as storage:
         count = 0
@@ -192,12 +244,12 @@ async def reverify(ctx):
     Reverify all NIP-05 metadata events
     """
     from .storage import get_storage
-    from .util import json
+    from .util import json_loads
 
     async with get_storage() as storage:
         count = 0
         async for event in storage.run_single_query([{"kinds": [0]}]):
-            meta = json.loads(event.content)
+            meta = json_loads(event.content)
             if "nip05" in meta:
                 try:
                     await storage.verifier.verify(event)
@@ -249,9 +301,7 @@ async def load(ctx, filename):
     else:
         fileobj = sys.stdin
     import collections
-    from .util import json
-
-    loads = json.loads
+    from .util import json_loads
 
     Config.authentication["enabled"] = False
     # this will reverify profiles but not reject unverified events
@@ -272,7 +322,7 @@ async def load(ctx, filename):
             line = fileobj.readline()
             if not line:
                 break
-            js = loads(line)
+            js = json_loads(line)
             if isinstance(js, list):
                 event = js[1]
             else:
@@ -402,17 +452,21 @@ def mirror(ids, authors, kinds, etags, ptags, since, until, limit, source, targe
 
 
 @main.command()
-@click.option("-q", "query", help="Query", required=True)
+@click.option("-q", "query", help="Query", required=False)
+@click.option("-a", "authors", type=click.File("r"), required=False)
 @click.pass_context
 @async_cmd
-async def purge(ctx, query):
+async def purge(ctx, query, authors):
     """
     Purge events from the specified query
     """
-    from .util import json
+    from .util import json_loads
     from .storage import get_storage
 
-    query = json.loads(query)
+    if authors:
+        query = {"authors": [a.strip() for a in authors.readlines()]}
+    else:
+        query = json_loads(query)
     print(repr(query))
     async with get_storage() as storage:
         events = []
@@ -454,3 +508,81 @@ def alembic(ctx, alembic_args):
         f'-c{os.path.join(os.path.dirname(__file__), "alembic.ini")}',
     ) + alembic_args
     return config.main(alembic_args)
+
+
+@click.group()
+@click.pass_context
+def fts(ctx):
+    """
+    Full Text Indexer Commands
+    """
+    pass
+
+
+@fts.command()
+@click.option("--since", type=int, required=False, default=1)
+@click.option("--until", type=int, required=False, default=0)
+@click.option("--batch", type=int, required=False, default=500)
+@async_cmd
+async def reindex(since, until, batch):
+    """
+    Reindex everything
+    """
+    import logging
+    import time
+
+    logging.basicConfig(
+        format="%(asctime)s %(name)s %(levelname)s %(message)s",
+        level=logging.INFO,
+    )
+    from nostr_relay.storage import get_storage
+
+    click.echo(f"Starting reindex at {int(time.time())}")
+    async with get_storage() as storage:
+        await storage.reindex("search", since=since, until=until, batch_size=batch)
+
+
+main.add_command(fts)
+
+
+@main.command()
+@async_cmd
+async def bench():
+    from nostr_relay.storage import get_storage
+    from nostr_relay.storage import kv
+    from nostr_relay.util import easy_profiler
+
+    import time
+
+    import logging
+
+    logging.basicConfig(level="INFO")
+
+    async with get_storage() as storage:
+        plan = kv.planner([{"kinds": [1], "limit": 50}])[0]
+        print(plan)
+
+        async def do_queries():
+            queries = 0
+            events = 0
+            start = time.perf_counter()
+            stop = start + 5
+
+            while time.perf_counter() < stop:
+                # for event in kv.execute_one_plan(storage.db, plan, storage.log)[1]:
+                #     events += 1
+                # queries += 1
+                async for event in storage.run_single_query(
+                    [{"kinds": [1], "limit": 25}]
+                ):
+                    events += 1
+                queries += 1
+
+            duration = time.perf_counter() - start
+            print(f"qps: {queries/duration:.1f}  events/sec: {events/duration:.1f}")
+
+        # with easy_profiler():
+        tasks = [asyncio.create_task(do_queries()) for i in range(3)]
+
+        # tasks = [asyncio.get_running_loop().run_in_executor(None, do_queries) for i in range(4)]
+        await asyncio.wait(tasks)
