@@ -2,7 +2,11 @@ import asyncio
 import logging
 import collections
 import functools
+import typing
+
 from aionostr.event import Event
+from pydantic import BaseModel, Field, validator, ValidationError
+
 from ..config import Config
 from ..errors import StorageError, AuthenticationError
 from ..util import (
@@ -57,7 +61,7 @@ class BaseStorage:
             self.notifier.start()
         else:
             self.notifier = None
-        self.stat_collector = StatsCollector(self.options.get("stats_interval", 60.0))
+        self.stat_collector = StatsCollector(self.options.pop("stats_interval", 60.0))
         await self.stat_collector.start()
         self.authenticator = get_authenticator(self, Config.get("authentication", {}))
         output_validator = Config.get("output_validator")
@@ -96,6 +100,14 @@ class BaseStorage:
             and len(self.clients[client_id]) == Config.subscription_limit
         ):
             raise StorageError("rejected: too many subscriptions")
+
+        try:
+            filters = [NostrQuery.parse_obj(q) for q in filters]
+        except ValidationError as e:
+            self.log.error(str(e))
+            await queue.put((sub_id, None))
+            return
+
         sub = self.subscription_class(
             self,
             sub_id,
@@ -113,6 +125,8 @@ class BaseStorage:
             sub.start()
             self.clients[client_id][sub_id] = sub
             self.log.debug("%s/%s +", client_id, sub_id)
+        else:
+            await queue.put((sub_id, None))
 
     async def unsubscribe(self, client_id, sub_id=None):
         if sub_id:
@@ -136,16 +150,16 @@ class BaseStorage:
 
     async def notify_other_processes(self, event):
         if self.notifier:
-            self._notifiy_sub_tasks.append(
+            self._notify_sub_tasks.append(
                 asyncio.create_task(self.notifier.notify(event))
             )
 
     async def notify_all_connected(self, event):
-        # notify all subscriptions
-        if self._notify_sub_tasks:
-            await asyncio.wait(self._notify_sub_tasks)
-            self._notify_sub_tasks.clear()
         with self.stat_collector.timeit("notify") as counter:
+            # notify all subscriptions
+            if self._notify_sub_tasks:
+                await asyncio.wait(self._notify_sub_tasks)
+                self._notify_sub_tasks.clear()
             for client in self.clients.values():
                 for sub in client.values():
                     self._notify_sub_tasks.append(
@@ -185,8 +199,7 @@ class BaseStorage:
     async def get_auth_roles(self, pubkey: str):
         query = {
             "kinds": [self.service_kind],
-            "#d": ["auth:roles"],
-            "#p": [pubkey],
+            "#d": [f"auth:{pubkey}"],
             "authors": [self.service_pubkey],
         }
         event = await self.get_event_from_query(query)
@@ -198,7 +211,7 @@ class BaseStorage:
     async def get_all_auth_roles(self):
         query = {
             "kinds": [self.service_kind],
-            "#d": ["auth:roles"],
+            "#t": ["auth"],
             "authors": [self.service_pubkey],
         }
         async for event in self.run_single_query([query]):
@@ -208,10 +221,58 @@ class BaseStorage:
                     yield tag[1], set((role or "").lower())
 
     async def set_auth_roles(self, pubkey: str, roles: str):
-        tags = {"p": pubkey, "d": "auth:roles"}
+        tags = {"t": "auth", "d": f"auth:{pubkey}", "p": pubkey}
         content = str(roles).lower()
         await self.add_service_event(content=content, tags=tags)
 
+    async def get_identified_pubkey(self, identifier="", domain=""):
+        query = {
+            "authors": [self.service_pubkey],
+            "kinds": [self.service_kind],
+            "#t": ["nip05"],
+        }
+
+        if identifier:
+            query["#n"] = [identifier]
+
+        data = {"names": {}, "relays": {}}
+        self.log.debug("Getting identity for %s %s", identifier, domain)
+        async for event in self.run_single_query([query]):
+            tags = collections.defaultdict(list)
+            for tag in event.tags:
+                tags[tag[0]].append(tag[1])
+            identifier, id_domain = tags["n"][0].split("@", 1)
+            if domain and id_domain != domain:
+                continue
+            data["names"][identifier] = event.content
+            if tags["r"]:
+                data["relays"][event.content] = tags["r"]
+
+        return data
+
+    async def set_identified_pubkey(self, identifier, pubkey, relays=None):
+        if pubkey:
+            tags = [
+                ["d", f"nip05:{pubkey}"],
+                ["t", "nip05"],
+                ["p", pubkey],
+                ["n", identifier],
+            ]
+            if relays:
+                for relay in relays:
+                    tags.append(["r", relay])
+            await self.add_service_event(tags=tags, content=pubkey)
+        else:
+            query = {
+                "#n": [identifier],
+                "#t": ["nip05"],
+                "authors": [self.service_pubkey],
+                "kinds": [self.service_kind],
+            }
+            event = await self.get_event_from_query(query)
+            if event:
+                await self.delete_event(event.id)
+
 
 class BaseSubscription:
     __slots__ = (
@@ -234,7 +295,7 @@ class BaseSubscription:
         filters: list,
         queue=None,
         client_id=None,
-        default_limit=6000,
+        default_limit=Config.max_limit,
         log=None,
         auth_token=None,
         **kwargs,
@@ -280,36 +341,87 @@ class BaseSubscription:
             await self.queue.put((self.sub_id, event))
 
     def check_event(self, event: Event, filters: list):
-        for filter_obj in filters:
-            if not filter_obj:
-                continue
+        for query in filters:
             matched = set()
-            for key, value in filter_obj.items():
-                if key == "ids":
-                    matched.add(event.id in value)
-                elif key == "authors":
-                    matched.add(event.pubkey in value)
-                    has_delegation, match = event.has_tag("delegation", value)
-                    if match:
-                        matched.add(True)
-                elif key == "kinds":
-                    matched.add(event.kind in value)
-                elif key == "since":
-                    matched.add(event.created_at >= value)
-                elif key == "until":
-                    matched.add(event.created_at < value)
-                elif key[0] == "#" and len(key) == 2:
-                    matched.add(all(event.has_tag(key[1], value)))
-                elif key == "limit":
-                    # limit is irrelevant for broadcasts
-                    continue
-                else:
-                    matched.add(False)
-            if all(matched):
+            if query.ids is not None:
+                matched.add(event.id in query.ids)
+            if query.authors is not None:
+                matched.add(event.pubkey in query.authors)
+                has_delegation, match = event.has_tag("delegation", query.authors)
+                if match:
+                    matched.add(True)
+            if query.kinds is not None:
+                matched.add(event.kind in query.kinds)
+            if query.since:
+                matched.add(event.created_at >= query.since)
+            if query.until:
+                matched.add(event.created_at < query.until)
+            if query.tags:
+                for tagname, values in query.tags:
+                    matched.add(all(event.has_tag(tagname, values)))
+            if matched and all(matched):
                 return True
         return False
 
 
+class NostrQuery(BaseModel):
+    """
+    Represents a valid nostr REQ query
+    """
+
+    ids: typing.Optional[list[str]] = Field(
+        anystr_lower=True, min_anystr_length=2, max_anystr_length=64
+    )
+    authors: typing.Optional[list[str]] = Field(
+        anystr_lower=True, min_anystr_length=2, max_anystr_length=64
+    )
+    kinds: typing.Optional[list[int]]
+    since: typing.Optional[int] = Field(ge=0, lt=2145934800)
+    until: typing.Optional[int] = Field(ge=0, lt=2145934800)
+    limit: typing.Optional[int] = Field(ge=0, default=Config.max_limit)
+    search: typing.Optional[str]
+    tags: typing.Optional[list[tuple[str, set]]]
+
+    @validator("ids", "authors", each_item=True)
+    def ids_are_hex(cls, hexid, field, **kwargs):
+        hexid = hexid.lower()
+        if any(i not in "abcdef0123456789" for i in hexid):
+            raise ValueError("not hex")
+        if len(hexid) < 2:
+            raise ValueError("too small")
+        return hexid
+
+    @validator("tags")
+    def check_tags(cls, values):
+        for tag, tagvalue in values:
+            for val in tagvalue:
+                if not isinstance(val, str):
+                    raise ValueError(f"{val} is not a string")
+        return values
+
+    @validator("ids", "authors", "kinds")
+    def sort_fields(cls, values):
+        return sorted(set(values), reverse=True)
+
+    @classmethod
+    def parse_obj(cls, obj):
+        if isinstance(obj, cls):
+            return obj
+        tags = []
+        try:
+            for k, v in obj.items():
+                if k.startswith("#") and len(k) == 2 and isinstance(v, list):
+                    tags.append((k[1], set(v)))
+            tags.sort(reverse=True)
+        except AttributeError as e:
+            raise StorageError("not a query")
+        if tags:
+            obj["tags"] = tags
+        else:
+            obj.pop("tags", None)
+        return super().parse_obj(obj)
+
+
 class BaseGarbageCollector(Periodic):
     def __init__(self, storage, **kwargs):
         self.log = logging.getLogger("nostr_relay.storage:gc")
