@@ -17,10 +17,11 @@ from ..util import (
     object_from_path,
     catchtime,
     Periodic,
-    json,
+    json_dumps,
+    json_loads,
 )
 from . import get_metadata
-from .base import BaseStorage, BaseSubscription, BaseGarbageCollector
+from .base import BaseStorage, BaseSubscription, BaseGarbageCollector, NostrQuery
 
 
 force_hex_translation = str.maketrans(
@@ -38,6 +39,21 @@ def validate_id(obj_id):
     return ""
 
 
+def event_from_tuple(row):
+    tags = row[4]
+    if isinstance(tags, str):
+        tags = json_loads(tags)
+    return Event(
+        id=row[0].hex(),
+        created_at=row[1],
+        kind=row[2],
+        pubkey=row[3].hex(),
+        tags=tags,
+        sig=row[5].hex(),
+        content=row[6],
+    )
+
+
 class DBStorage(BaseStorage):
     DEFAULT_GARBAGE_COLLECTOR = "nostr_relay.storage.db.QueryGarbageCollector"
 
@@ -120,8 +136,8 @@ class DBStorage(BaseStorage):
 
         self.db = create_async_engine(
             self.db_url,
-            json_deserializer=json.loads,
-            json_serializer=json.dumps,
+            json_deserializer=json_loads,
+            json_serializer=json_dumps,
             pool_pre_ping=True,
             **self.sqlalchemy_options,
         )
@@ -164,7 +180,7 @@ class DBStorage(BaseStorage):
                 )
                 row = result.first()
         if row:
-            return Event.from_tuple(row)
+            return event_from_tuple(row)
 
     async def delete_event(self, event_id):
         async with self.db.begin() as conn:
@@ -339,13 +355,14 @@ class DBStorage(BaseStorage):
         """
         Run a single query, yielding json events
         """
+        nostr_queries = [NostrQuery.parse_obj(q) for q in query_filters]
         queue = asyncio.Queue()
         sub = self.subscription_class(
-            self, "", query_filters, queue=queue, default_limit=600000
+            self, "", nostr_queries, queue=queue, default_limit=600000
         )
-        sub.prepare()
-        async for event in self.run_query(sub.query):
-            yield event
+        if sub.prepare():
+            async for event in self.run_query(sub.query):
+                yield event
 
     async def run_query(self, query, if_long=None):
         self.log.debug(query)
@@ -355,7 +372,7 @@ class DBStorage(BaseStorage):
                     async with self.db.connect() as conn:
                         async with conn.stream(query) as result:
                             async for row in result:
-                                yield Event.from_tuple(row)
+                                yield event_from_tuple(row)
                                 counter["count"] += 1
             duration = counter["duration"]
             if duration > 1.0 and if_long:
@@ -449,7 +466,7 @@ class DBStorage(BaseStorage):
             elif not (validate_id(pubkey) and len(pubkey) == 64):
                 raise StorageError("invalid public key")
             else:
-                pars = [identifier, pubkey, json.dumps(relays or [])]
+                pars = [identifier, pubkey, json_dumps(relays or [])]
                 await conn.execute(
                     sa.delete(self.IdentTable).where(
                         self.IdentTable.c.identifier == identifier
@@ -536,7 +553,7 @@ class Subscription(BaseSubscription):
                 "nostr_relay.long-queries"
             ).warning(
                 f"{self.client_id}/{self.sub_id} Long query: '{self.filters}' took %dms",
-                duration,
+                duration * 1000,
             ),
         )
         if check_output:
@@ -554,83 +571,69 @@ class Subscription(BaseSubscription):
         await queue.put((sub_id, None))
 
     def evaluate_filter(self, filter_obj, subwhere):
-        for key, value in filter_obj.items():
-            if key == "ids":
-                if not isinstance(value, list):
-                    value = [value]
-                ids = set(value)
-                if ids:
-                    exact = []
-                    for eid in ids:
-                        eid = validate_id(eid)
-                        if eid:
-                            if len(eid) == 64:
-                                if self.is_postgres:
-                                    exact.append(f"'\\x{eid}'")
-                                else:
-                                    exact.append(f"x'{eid}'")
-                            elif len(eid) > 2:
-                                if self.is_postgres:
-                                    subwhere.append(f"encode(id, 'hex') LIKE '{eid}%'")
-                                else:
-                                    subwhere.append(f"lower(hex(id)) LIKE '{eid}%'")
-                    if exact:
-                        idstr = ",".join(exact)
-                        subwhere.append(f"events.id IN ({idstr})")
-                else:
-                    # invalid query
-                    raise ValueError("ids")
-            elif key == "authors" and isinstance(value, list):
-                if value:
-                    exact = set()
-                    hexexact = set()
-                    for pubkey in value:
-                        pubkey = validate_id(pubkey)
-                        if pubkey:
-                            if len(pubkey) == 64:
-                                if self.is_postgres:
-                                    exact.add(f"'\\x{pubkey}'")
-                                else:
-                                    exact.add(f"x'{pubkey}'")
-                                hexexact.add(f"'{pubkey}'")
-                            # no prefix searches, for now
-                    if exact:
-                        astr = ",".join(exact)
-                        subwhere.append(
-                            f"(pubkey IN ({astr}) OR id IN (SELECT id FROM tags WHERE name = 'delegation' AND value IN ({','.join(hexexact)})))"
-                        )
-                    else:
-                        raise ValueError("authors")
-                else:
-                    # query with empty list should be invalid
-                    raise ValueError("authors")
-            elif key == "kinds":
-                if isinstance(value, list) and all(isinstance(k, int) for k in value):
+        if filter_obj.ids is not None:
+            if filter_obj.ids:
+                exact = []
+                for eid in filter_obj.ids:
+                    if len(eid) == 64:
+                        if self.is_postgres:
+                            exact.append(f"'\\x{eid}'")
+                        else:
+                            exact.append(f"x'{eid}'")
+                    elif len(eid) > 2:
+                        if self.is_postgres:
+                            subwhere.append(f"encode(id, 'hex') LIKE '{eid}%'")
+                        else:
+                            subwhere.append(f"lower(hex(id)) LIKE '{eid}%'")
+                if exact:
+                    idstr = ",".join(exact)
+                    subwhere.append(f"events.id IN ({idstr})")
+            else:
+                raise ValueError("ids")
+        if filter_obj.authors is not None:
+            if filter_obj.authors:
+                exact = set()
+                hexexact = set()
+                for pubkey in filter_obj.authors:
+                    if len(pubkey) == 64:
+                        if self.is_postgres:
+                            exact.add(f"'\\x{pubkey}'")
+                        else:
+                            exact.add(f"x'{pubkey}'")
+                        hexexact.add(f"'{pubkey}'")
+                        # no prefix searches, for now
+                if exact:
+                    astr = ",".join(exact)
                     subwhere.append(
-                        "kind IN ({})".format(",".join(str(int(k)) for k in value))
+                        f"(pubkey IN ({astr}) OR id IN (SELECT id FROM tags WHERE name = 'delegation' AND value IN ({','.join(hexexact)})))"
                     )
                 else:
-                    raise ValueError("kinds")
-            elif key == "since":
-                if value:
-                    subwhere.append("created_at >= %d" % int(value))
-            elif key == "until":
-                if value:
-                    subwhere.append("created_at < %d" % int(value))
-                else:
-                    raise ValueError("until")
-            elif key == "limit" and isinstance(value, int):
-                filter_obj["limit"] = min(max(value, 0), self.default_limit)
-            elif key[0] == "#" and len(key) == 2 and value:
+                    raise ValueError("authors")
+            else:
+                # query with empty list should be invalid
+                raise ValueError("authors")
+        if filter_obj.kinds is not None:
+            if filter_obj.kinds:
+                subwhere.append(
+                    "kind IN ({})".format(",".join(str(k) for k in filter_obj.kinds))
+                )
+            else:
+                raise ValueError("kinds")
+        if filter_obj.since is not None:
+            subwhere.append("created_at >= %d" % filter_obj.since)
+        if filter_obj.until is not None:
+            subwhere.append("created_at < %d" % filter_obj.until)
+        if filter_obj.tags:
+            for tagname, tags in filter_obj.tags:
                 pstr = []
-                for val in set(value):
+                for val in tags:
                     if val:
                         val = val.replace("'", "''")
                         pstr.append(f"'{val}'")
                 if pstr:
                     pstr = ",".join(pstr)
                     subwhere.append(
-                        f"id IN (SELECT id FROM tags WHERE name = '{key[1]}' AND value IN ({pstr})) "
+                        f"id IN (SELECT id FROM tags WHERE name = '{tagname}' AND value IN ({pstr})) "
                     )
         return filter_obj
 
@@ -647,15 +650,15 @@ class Subscription(BaseSubscription):
                 filter_obj = self.evaluate_filter(filter_obj, subwhere)
             except ValueError:
                 self.log.debug("bad query %s", filter_obj)
-                filter_obj = {}
+                filter_obj = NostrQuery()
                 subwhere = []
             if subwhere:
                 subwhere = " AND ".join(subwhere)
                 where.add(subwhere)
             else:
                 where.add("false")
-            if "limit" in filter_obj:
-                limit = filter_obj["limit"]
+            if filter_obj.limit:
+                limit = min(filter_obj.limit, Config.max_limit)
             new_filters.append(filter_obj)
         if where:
             select += " WHERE (\n\t"
