@@ -9,20 +9,27 @@ import collections
 import threading
 import functools
 import traceback
+import re
+import os.path
 
 from contextlib import contextmanager
 from queue import SimpleQueue
-from time import perf_counter, time
+from time import perf_counter, time, sleep
 
 from aionostr.event import Event, EventKind
 from msgpack import packb, unpackb
 from concurrent import futures
 
-from .base import BaseStorage, BaseSubscription, BaseGarbageCollector
+from .base import (
+    BaseStorage,
+    BaseSubscription,
+    BaseGarbageCollector,
+    NostrQuery,
+    ValidationError,
+)
 from ..auth import get_authenticator
 from ..config import Config
 from ..errors import StorageError
-from ..util import catchtime, Periodic, StatsCollector
 from ..validators import get_validator
 
 
@@ -34,8 +41,6 @@ from ..validators import get_validator
 
 DONE = object()
 
-QUERY_POOL = futures.ThreadPoolExecutor(max_workers=20)
-
 
 VERSION = 1
 
@@ -54,6 +59,12 @@ QueryPlan = collections.namedtuple(
 )
 
 
+class QueryPlans(list):
+    """
+    Container for QueryPlan objects
+    """
+
+
 class FakeContainer:
     def __contains__(self, anything):
         return True
@@ -61,6 +72,7 @@ class FakeContainer:
 
 class Index:
     cardinality = 1
+    enabled = True
 
     def __init__(self):
         self.hits = self.misses = 0
@@ -76,6 +88,11 @@ class Index:
     def clear(self, event, txn):
         self.write(event, txn, operation="delete")
 
+    def bulk_update(self, events, txn):
+        for event in events:
+            if event:
+                self.write(event, txn)
+
     def __repr__(self):
         return f"{self.__class__.__name__}({self.prefix})"
 
@@ -88,11 +105,14 @@ class Index:
         # to make substring checks quicker
         compiled_matches = []
         for match in sorted(matches, reverse=True):
-            compiled_matches.append(self.to_key(match))
+            try:
+                compiled_matches.append(self.to_key(match))
+            except ValueError:
+                pass
         if since is not None:
-            since = max(int(since), 1).to_bytes(4, "big")
+            since = since.to_bytes(4, "big")
         if until is not None:
-            until = max(int(until), 1).to_bytes(4, "big")
+            until = until.to_bytes(4, "big")
             add_time = b"\x00%s\x00" % until
         else:
             add_time = b""
@@ -107,7 +127,7 @@ class Index:
                 prev()
             return found
 
-        if matches:
+        if compiled_matches:
             next_match = iter(compiled_matches).__next__
             match = next_match()
             if len(compiled_matches) > 1:
@@ -115,7 +135,7 @@ class Index:
             else:
                 stop = self.prefix
             if since:
-                stop += b"\x00%s" % since
+                stop += b"\x00" + since
             skip(match + add_time)
         else:
             match = None
@@ -160,8 +180,9 @@ class Index:
                     elif key < stop:
                         break
 
-                    if key[-32:] in events:
-                        yield key[-32:]
+                    event_id = key[-32:]
+                    if event_id in events:
+                        yield event_id
                     if not prev():
                         break
                     key = bytes(get_key())
@@ -188,7 +209,7 @@ class IdIndex(Index):
     cardinality = 1000
 
     def to_key(self, value):
-        return self.prefix + bytes.fromhex(value)
+        return self.prefix + bytes_from_hex(value)
 
     def write(self, event, txn, operation="put"):
         if operation == "put":
@@ -221,7 +242,7 @@ class PubkeyIndex(Index):
     prefix = b"\x03"
 
     def to_key(self, value):
-        return self.prefix + bytes.fromhex(value)
+        return self.prefix + bytes_from_hex(value)
 
     def convert(self, event):
         yield self.to_key(event.pubkey)
@@ -248,7 +269,7 @@ class AuthorKindIndex(Index):
     def to_key(self, value):
         return b"%s%s\x00%s" % (
             self.prefix,
-            bytes.fromhex(value[0]),
+            bytes_from_hex(value[0]),
             value[1].to_bytes(4, "big"),
         )
 
@@ -256,6 +277,109 @@ class AuthorKindIndex(Index):
         yield self.to_key((event.pubkey, event.kind))
 
 
+class FTSIndex(Index):
+    """
+    Full-text index, using whoosh
+    """
+
+    cardinality = 1000
+    prefix = b""
+
+    def __init__(self, index_path=Config.fts_index_path, enabled=Config.fts_enabled):
+        self.index_path = index_path
+        self._schema = None
+        self._index = None
+
+        self.log = logging.getLogger("nostr_relay.fts")
+        self.enabled = Config.fts_enabled
+
+    @property
+    def schema(self):
+        if self._schema is None:
+            from whoosh.fields import Schema, TEXT, ID, NUMERIC
+            from whoosh.analysis import StemmingAnalyzer
+
+            self._schema = Schema(
+                id=ID(stored=True, unique=True),
+                created_at=NUMERIC(stored=True),
+                content=TEXT(analyzer=StemmingAnalyzer()),
+            )
+        return self._schema
+
+    @property
+    def index(self):
+        if self._index is None:
+            storage_path = self.index_path or os.path.join(
+                Config.storage["path"], "fts"
+            )
+            from whoosh.filedb.filestore import FileStorage
+            from whoosh.index import EmptyIndexError
+
+            storage = FileStorage(storage_path)
+            storage.create()
+            try:
+                self._index = storage.open_index()
+            except EmptyIndexError:
+                storage.create()
+                self._index = storage.create_index(self.schema)
+        return self._index
+
+    def _index_item(self, writer, event):
+        writer.update_document(
+            id=event.id, content=event.content, created_at=event.created_at
+        )
+        self.log.debug("Indexed %s", event.id)
+
+    def write(self, event, txn):
+        if event.kind in (1, 0):
+            from whoosh.writing import AsyncWriter
+
+            with AsyncWriter(self.index) as writer:
+                self._index_item(writer, event)
+
+    def bulk_update(self, events, txn):
+        from whoosh.writing import BufferedWriter
+
+        with BufferedWriter(self.index, limit=200) as writer:
+            for event in events:
+                if event:
+                    self._index_item(writer, event)
+            writer.commit()
+
+    def clear(self, event, txn):
+        with self.index.writer() as writer:
+            writer.delete_by_term("id", event.id)
+
+    @contextmanager
+    def scanner(
+        self,
+        txn,
+        matches: list,
+        since=None,
+        until=None,
+        events=FakeContainer(),
+        limit=500,
+    ):
+        from whoosh.query import NumericRange
+        from whoosh.qparser import QueryParser
+
+        qp = QueryParser("content", schema=self.schema)
+        nr = NumericRange("created_at", since or 1, until or int(time()))
+
+        def iterator():
+            with self.index.searcher() as searcher:
+                for query in matches:
+                    q = qp.parse(query) & nr
+                    results = searcher.search(q, limit=limit)
+
+                    for result in results:
+                        event_id = bytes.fromhex(result["id"])
+                        if event_id in events:
+                            yield event_id
+
+        yield iterator()
+
+
 INDEXES = {
     "ids": IdIndex(),
     "created_at": CreatedIndex(),
@@ -263,6 +387,7 @@ INDEXES = {
     "authors": PubkeyIndex(),
     "authorkinds": AuthorKindIndex(),
     "tags": TagIndex(),
+    "search": FTSIndex(),
 }
 
 
@@ -328,52 +453,46 @@ class WriterThread(threading.Thread):
         self.env = env
         self.stat_collector = stat_collector
         self.queue = SimpleQueue()
+        self.write_indexes = [i for i in INDEXES.values() if i.enabled]
+        self.processing = False
 
     def run(self):
         env = self.env
         qget = self.queue.get
         qsize = self.queue.qsize
         stat_collector = self.stat_collector
-        write_indexes = INDEXES.values()
         log = logging.getLogger("nostr_relay.writer")
 
-        def delete_event(event_id, txn):
-            event_data = txn.get(b"\x00%s" % event_id)
-            if event_data:
-                event = decode_event(event_data)
-                for index in reversed(write_indexes):
-                    index.clear(event, txn)
-
         while self.running:
             task = qget()
+            self.processing = True
             if task is None:
                 break
-            event, operation = task
+            operation, args = task
             try:
                 with stat_collector.timeit("write") as counter:
                     with env.begin(write=True, buffers=True) as txn:
-                        if operation == "add":
-                            for index in write_indexes:
+                        if operation == "add" and not get_event_data(
+                            txn, args[0].id_bytes
+                        ):
+                            event = args[0]
+                            for index in self.write_indexes:
                                 index.write(event, txn)
                                 # log.debug("index %s event %s", index, event)
-                            counter["count"] += 1
-                            if event.kind in (
-                                EventKind.SET_METADATA,
-                                EventKind.CONTACTS,
-                            ):
-                                # delete older metadata events
-                                with INDEXES["authorkinds"].scanner(
-                                    txn,
-                                    [(event.pubkey, event.kind)],
-                                    until=event.created_at - 1,
-                                ) as scanner:
-                                    for event_id in scanner:
-                                        log.info("Deleting %s", event_id)
-                                        delete_event(event_id, txn)
-                                        counter["count"] += 1
+                            self._post_save(txn, event, counter, log)
                         elif operation == "del":
-                            delete_event(bytes.fromhex(event), txn)
-                            counter["count"] += 1
+                            event = decode_event(
+                                get_event_data(txn, bytes.fromhex(args[0]))
+                            )
+                            if event:
+                                self._delete_event(txn, event, log)
+                        elif operation == "reindex":
+                            index_name, event = args
+                            INDEXES[index_name].write(event, txn)
+                        elif operation == "bulk_update":
+                            index_name, events = args
+                            INDEXES[index_name].bulk_update(events, txn)
+                        counter["count"] += 1
                 qs = qsize()
                 if qs >= 1000 and qs % 1000 == 0:
                     # since we can do about 1,000 writes per second (end-to-end),
@@ -381,17 +500,84 @@ class WriterThread(threading.Thread):
                     log.warning("Write queue size: %d", qs)
             except Exception as e:
                 log.exception("writer")
+            finally:
+                self.processing = False
+
+    def _delete_event(self, txn, event, log):
+        for index in reversed(self.write_indexes):
+            index.clear(event, txn)
+        log.info(
+            "Deleted event %s kind=%d pubkey=%s", event.id, event.kind, event.pubkey
+        )
+
+    def _post_save(self, txn, event, counter, log):
+        if (
+            event.kind
+            in (
+                EventKind.SET_METADATA,
+                EventKind.CONTACTS,
+            )
+            or event.is_replaceable
+            or event.is_paramaterized_replaceable
+        ):
+            saved_id = event.id_bytes
+            until = event.created_at - 1
+            if event.is_paramaterized_replaceable:
+                d_tag = [tag[1] for tag in event.tags if tag[0] == "d"][0]
+            else:
+                d_tag = None
+
+            # delete older replaceable events
+            with INDEXES["authorkinds"].scanner(
+                txn,
+                [(event.pubkey, event.kind)],
+                until=until,
+            ) as scanner:
+                for event_id in scanner:
+                    if event_id == saved_id:
+                        continue
+                    candidate = decode_event(get_event_data(txn, event_id))
+                    if d_tag is not None:
+                        if not all(candidate.has_tag("d", d_tag)):
+                            continue
+                    self._delete_event(txn, candidate, log)
+                    counter["count"] += 1
+
+        elif event.kind == EventKind.DELETE:
+            # delete the referenced events
+            ids = set((bytes_from_hex(tag[1]) for tag in event.tags if tag[0] == "e"))
+            if not ids:
+                return
+            with INDEXES["authors"].scanner(
+                txn,
+                [event.pubkey],
+                until=event.created_at - 1,
+            ) as scanner:
+                for event_id in scanner:
+                    if event_id in ids:
+                        candidate = decode_event(get_event_data(txn, event_id))
+                        if candidate:
+                            self._delete_event(txn, candidate, log)
+                            counter["count"] += 1
 
 
 class LMDBStorage(BaseStorage):
+    """
+    LMDB Storage backend
+    """
+
     DEFAULT_GARBAGE_COLLECTOR = "nostr_relay.storage.kv.KVGarbageCollector"
 
     def __init__(self, options):
         super().__init__(options)
+        pool_size = self.options.pop("pool_size", 8)
         self.options.pop("class")
+        # set this to the number of read threads
+        self.options.setdefault("max_spare_txns", pool_size)
         self.log = logging.getLogger(__name__)
         self.subscription_class = Subscription
         self.db = None
+        self.query_pool = futures.ThreadPoolExecutor(max_workers=pool_size)
 
     async def close(self):
         if self.db:
@@ -399,6 +585,7 @@ class LMDBStorage(BaseStorage):
                 self.garbage_collector_task.cancel()
             self.writer_queue.put(None)
             self.writer_thread.join()
+            self.query_pool.shutdown()
             self.db.close()
             self.db = None
             self.log.debug("Closed storage %s", self.options["path"])
@@ -414,6 +601,15 @@ class LMDBStorage(BaseStorage):
         self.writer_queue = self.writer_thread.queue
         self.writer_thread.start()
 
+    async def wait_for_writer(self):
+        """
+        Wait for the writer thread to be idle.
+        ONLY USE THIS FOR TESTING
+        """
+        await asyncio.sleep(0.1)
+        while self.writer_thread.processing or not self.writer_queue.empty():
+            await asyncio.sleep(0.1)
+
     def write_tombstone(self):
         """
         Write a record at the end of the db, to avoid a weird edge case
@@ -421,8 +617,8 @@ class LMDBStorage(BaseStorage):
         with self.db.begin(write=True) as txn:
             txn.put(b"\xee", b"")
 
-    def delete_event(self, event_id):
-        self.writer_queue.put((event_id, "del"))
+    async def delete_event(self, event_id):
+        self.writer_queue.put(("del", [event_id]))
 
     async def add_event(self, event_json: dict, auth_token=None):
         """
@@ -437,68 +633,68 @@ class LMDBStorage(BaseStorage):
 
         await self.validate_event(event, Config)
 
-        self.writer_queue.put((event, "add"))
+        self.writer_queue.put(("add", [event]))
         await self.post_save(event)
         return event, True
 
     async def post_save(self, event, **kwargs):
-        query = None
-        if event.is_replaceable or event.is_paramaterized_replaceable:
-            query = {
-                "authors": [event.pubkey],
-                "until": event.created_at - 1,
-                "kinds": [event.kind],
-            }
-            if event.is_paramaterized_replaceable:
-                query["#d"] = [[tag[1] for tag in event.tags if tag[0] == "d"][0]]
-            self.log.debug(query)
-        elif event.kind == EventKind.DELETE:
-            # delete the referenced events
-            ids = []
-            for tag in event.tags:
-                name = tag[0]
-                if name == "e":
-                    ids.append(tag[1])
-            query = {
-                "authors": [event.pubkey],
-                "ids": ids,
-                "until": event.created_at - 1,
-            }
-        if query:
-            task, events = executor(self.db, [query], loop=self.loop)
-            await task
-            for event in events:
-                if event is not None:
-                    self.delete_event(event.id)
-                    self.log.info(
-                        "Deleted/Replaced event %s kind=%d pubkey=%s",
-                        event.id,
-                        event.kind,
-                        event.pubkey,
-                    )
-
         await self.notify_all_connected(event)
         # notify other processes
         await self.notify_other_processes(event)
 
+    async def reindex(self, index_name, batch_size=500, kinds=(1, 0), since=1, until=0):
+        query = {
+            "kinds": kinds,
+            "limit": batch_size,
+            "until": until or int(time()) + 10,
+            "since": since,
+        }
+        self.log.info(
+            "Starting reindex for %s from %d to %d",
+            index_name,
+            query["since"],
+            query["until"],
+        )
+        full_count = 0
+        while True:
+            async for plan, events in executor(
+                self.db, [query], self.query_pool, default_limit=batch_size
+            ):
+                while self.writer_queue.qsize() >= 1:
+                    await asyncio.sleep(2.0)
+                self.writer_queue.put(("bulk_update", [index_name, events]))
+                full_count += len(events)
+                self.log.info(
+                    "Bulk updating %s index. batch size: %d/%d count: %d",
+                    index_name,
+                    len(events),
+                    batch_size,
+                    full_count,
+                )
+            if len(events) == batch_size:
+                query["until"] = events[-1].created_at + 2
+            else:
+                break
+        self.log.info(
+            "Done bulk updating %s. %d events indexed", index_name, full_count
+        )
+
     async def get_event(self, event_id: str):
         with self.db.begin(buffers=True) as txn:
-            event_data = txn.get(b"\x00%s" % bytes.fromhex(event_id))
-            if event_data:
-                return decode_event(event_data)
+            return decode_event(get_event_data(txn, bytes.fromhex(event_id)))
 
     async def run_single_query(self, filters):
         if isinstance(filters, dict):
             filters = [filters]
-        task, events = executor(self.db, filters, default_limit=600000)
-        await task
-        for event in events:
-            if event is not None:
-                yield event
-
-    async def get_identified_pubkey(self, identifier, domain=""):
-        data = {"names": {}, "relays": {}}
-        return data
+        plans = []
+        async for plan, events in executor(
+            self.db, filters, self.query_pool, default_limit=600000, log=self.log
+        ):
+            for event in events:
+                if event is not None:
+                    yield event
+            plans.append(plan)
+        analyze(plans, log=self.log)
 
     async def get_stats(self):
         stats = {}
@@ -515,11 +711,6 @@ class LMDBStorage(BaseStorage):
 
 
 class Subscription(BaseSubscription):
-    __slots__ = (
-        "filter_json",
-        # "check_event"
-    )
-
     def prepare(self):
         self.query = planner(
             self.filters, default_limit=self.default_limit, log=self.log
@@ -529,19 +720,17 @@ class Subscription(BaseSubscription):
     async def run_query(self):
         check_output = self.storage.check_output
         sub_id = self.sub_id
-        queue = self.queue
+        queue_put = self.queue.put
         with self.storage.stat_collector.timeit("query") as counter:
-            task, events = executor(
+            plans = []
+            iterator = executor(
                 self.storage.db,
                 self.query,
-                loop=self.storage.loop,
+                self.storage.query_pool,
                 default_limit=self.default_limit,
                 log=self.log,
+                loop=self.storage.loop,
             )
-
-            # we could start consuming events before all queries are complete,
-            # but it creates more complexity and is actually slower because of thread contention
-            await task
             try:
                 if check_output:
                     context = {
@@ -549,18 +738,27 @@ class Subscription(BaseSubscription):
                         "client_id": self.client_id,
                         "auth_token": self.auth_token,
                     }
-                    for event in events:
-                        if check_output(event, context):
-                            await queue.put((sub_id, event))
-                            counter["count"] += 1
+                    async for plan, events in iterator:
+                        for event in events:
+                            if check_output(event, context):
+                                await queue_put((sub_id, event))
+                                counter["count"] += 1
+                        plans.append(plan)
                 else:
-                    for event in events:
-                        await queue.put((sub_id, event))
-                        counter["count"] += 1
+                    async for plan, events in iterator:
+                        for event in events:
+                            await queue_put((sub_id, event))
+                            counter["count"] += 1
+                        plans.append(plan)
+            except asyncio.exceptions.CancelledError:
+                # cancellations are normal
+                self.log.debug("Cancelled run_query")
             except:
                 self.log.exception("run_query")
             finally:
-                await queue.put((sub_id, None))
+                await queue_put((sub_id, None))
+                analyze(plans)
+
         self.log.debug("Done with query")
 
 
@@ -597,67 +795,88 @@ class KVGarbageCollector(BaseGarbageCollector):
         cursor.close()
         if to_del:
             for event_id in to_del:
-                self.storage.delete_event(event_id)
+                await self.storage.delete_event(event_id)
         return len(to_del)
 
 
-def planner(filters, default_limit=6000, log=None):
+def planner(filters, default_limit=6000, log=None, maximum_plans=5):
     """
     Create a list of QueryPlans for the the list of REQ filters
     """
-    plans = []
-    for query in filters:
+    plans = QueryPlans()
+    for query in filters[:maximum_plans]:
         if isinstance(query, QueryPlan):
             plans.append(query)
             continue
-        if not query:
-            if log:
-                log.info("No empty queries allowed")
-            continue
-        tags = set()
+        elif not isinstance(query, NostrQuery):
+            if not query:
+                if log:
+                    log.info("No empty queries allowed")
+                continue
+            try:
+                query = NostrQuery.parse_obj(query)
+            except ValidationError as e:
+                if log:
+                    log.info("invalid query %s", e)
+                continue
+
         query_items = []
 
-        since = query.pop("since", None)
-        if since is not None:
-            query_items.append(("since", since))
-        until = query.pop("until", None)
-        if until is not None:
-            query_items.append(("until", until))
-        try:
-            limit = min(max(0, int(query.pop("limit"))), default_limit)
-        except KeyError:
-            limit = default_limit
+        if query.since is not None:
+            query_items.append(("since", query.since))
+
+        if query.until is not None:
+            query_items.append(("until", query.until))
 
         best_index = MultiIndex()
-        if "ids" in query:
-            ids = query.pop("ids")
-            query_items.append(("ids", tuple(ids)))
-            best_index.add("ids", ids)
-        if "kinds" in query and "authors" in query:
-            kinds = tuple(sorted(query.pop("kinds"), reverse=True))
-            authors = tuple(sorted(query.pop("authors"), reverse=True))
+        has_authors = has_kinds = False
+        if query.ids is not None:
+            ids = tuple(query.ids)
+            if ids:
+                query_items.append(("ids", ids))
+                best_index.add("ids", ids)
+            else:
+                continue
+        if query.kinds is not None:
+            kinds = tuple(query.kinds)
+            if kinds:
+                query_items.append(("kinds", kinds))
+                has_kinds = True
+            else:
+                continue
+        if query.authors is not None:
+            authors = tuple(query.authors)
+            if authors:
+                query_items.append(("authors", authors))
+                has_authors = True
+            else:
+                continue
+
+        if has_kinds and has_authors:
             authormatches = []
             for author in authors:
                 for k in kinds:
                     authormatches.append((author, k))
-            query_items.append(("kinds", kinds))
-            query_items.append(("authors", authors))
             best_index.add("authorkinds", authormatches)
-        elif "kinds" in query:
-            kinds = tuple(sorted(query.pop("kinds"), reverse=True))
-            query_items.append(("kinds", kinds))
+        elif has_kinds:
             best_index.add("kinds", kinds)
-        elif "authors" in query:
-            authors = tuple(sorted(query.pop("authors"), reverse=True))
-            query_items.append(("authors", authors))
+        elif has_authors:
             best_index.add("authors", authors)
 
-        for key, value in query.items():
-            if key[0] == "#" and len(key) == 2 and isinstance(value, list):
-                tag = key[1]
-                tags.update((tag, str(val)) for val in value)
-                query_items.append((key, tuple(value)))
-        if tags:
+        if query.search is not None and Config.fts_enabled:
+            # NIP-50 specifies name:value as extensions
+            # let's just remove them from the query
+            search_term = re.sub(r"([\w]+:[\w]+)", "", query.search).strip()
+            if len(search_term) > 3:
+                query_items.append(("search", search_term))
+                best_index.add("search", (search_term,))
+
+        if query.tags:
+            tags = set()
+            for tag, values in query.tags:
+                for val in values:
+                    tags.add((tag, val))
+                query_items.append((tag, tuple(values)))
             best_index.add("tags", sorted(tags, reverse=True))
 
         query_items = tuple(query_items)
@@ -665,7 +884,7 @@ def planner(filters, default_limit=6000, log=None):
         # set the final order of the multiindex
         best_index, matches = best_index.finalize()
 
-        if best_index is INDEXES["created_at"] and not (since or until):
+        if best_index is INDEXES["created_at"] and not (query.since or query.until):
             # don't allow range scans
             if log:
                 log.info("No range scans allowed %s", query_items)
@@ -675,9 +894,9 @@ def planner(filters, default_limit=6000, log=None):
             query_items,
             best_index,
             matches,
-            limit,
-            since,
-            until,
+            min(query.limit, default_limit),
+            query.since,
+            query.until,
             {},
         )
         if log:
@@ -693,16 +912,10 @@ def matcher(txn, event_id_iterator, query_items: tuple, stats: dict):
     """
     match = compile_match_from_query(query_items)
 
-    def get_event_data(event_id):
-        try:
-            return unpackb(txn.get(b"\x00%s" % event_id))
-        except TypeError:
-            return None
-
     stats["index_hits"] = stats["index_misses"] = 0
     for event_id in event_id_iterator:
-        event_tuple = get_event_data(event_id)
-        if match(event_tuple):
+        event_tuple = get_event_data(txn, event_id)
+        if event_tuple and match(event_tuple):
             event = Event(
                 id=event_tuple[1].hex(),
                 created_at=event_tuple[2],
@@ -724,10 +937,20 @@ def compile_match_from_query(query_items):
     for key, value in query_items:
         if key == "ids":
             col = FIELDS_TO_COLUMNS["id"]
-            filter_clauses.add(f"(et[{col}].hex() in {value!r})")
+            if all(len(v) == 64 for v in value):
+                filter_clauses.add(f"(et[{col}].hex() in {value!r})")
+            else:
+                filter_clauses.add(
+                    f"any(et[{col}].hex().startswith(v) for v in {value!r})"
+                )
         elif key == "authors":
             col = FIELDS_TO_COLUMNS["pubkey"]
-            filter_clauses.add(f"(et[{col}].hex() in {value!r})")
+            if all(len(v) == 64 for v in value):
+                filter_clauses.add(f"(et[{col}].hex() in {value!r})")
+            else:
+                filter_clauses.add(
+                    f"any(et[{col}].hex().startswith(v) for v in {value!r})"
+                )
         elif key == "kinds":
             col = FIELDS_TO_COLUMNS["kind"]
             filter_clauses.add(f"(et[{col}] in {value!r})")
@@ -737,11 +960,15 @@ def compile_match_from_query(query_items):
         elif key == "until" and value:
             col = FIELDS_TO_COLUMNS["created_at"]
             filter_clauses.add(f"(et[{col}] <= {value!r})")
-        elif key[0] == "#" and len(key) == 2:
+        elif key == "search" and Config.fts_enabled:
+            # assume that the index matched correctly
+            filter_clauses.add("True")
+        else:
             col = FIELDS_TO_COLUMNS["tags"]
             filter_clauses.add(
-                f"bool([t for t in et[{col}] if t[0] == {key[1]!r} and len(t) > 1 and t[1] in {value!r}])"
+                f"bool([t for t in et[{col}] if t[0] == {key!r} and len(t) > 1 and t[1] in {value!r}])"
             )
+
     filter_string = " and ".join(filter_clauses)
     function = f"""
 def check(et):
@@ -756,58 +983,48 @@ def check(et):
     return loc["check"]
 
 
-def executor(
+async def executor(
     lmdb_environment: lmdb.Environment,
-    filters: list,
-    loop=None,
+    plans: list,
+    pool,
     default_limit=6000,
     log=None,
+    loop=None,
 ):
     """
     Execute one or several queries.
-    Returns (future, event_list)
-    await the future before accessing the list
+    Returns iterator of (plan, events)
     """
     log = log or logging.getLogger("nostr_relay.kvquery")
 
-    plans = planner(filters, default_limit=default_limit, log=log)
-    loop = loop or asyncio.get_running_loop()
-
-    try:
-        # collections.deque is threadsafe for appends and pops
-        # and faster than using a Queue
-        events = collections.deque()
-
-        tasks = []
-        for plan in plans:
-            tasks.append(
-                loop.run_in_executor(
-                    QUERY_POOL,
-                    execute_one_plan,
-                    lmdb_environment,
-                    plan,
-                    events.append,
-                    log,
-                )
+    if not isinstance(plans, QueryPlans):
+        plans = planner(plans, default_limit=default_limit, log=log)
+    if len(plans) == 1:
+        fut = asyncio.wrap_future(
+            pool.submit(execute_one_plan, lmdb_environment, plans[0], log), loop=loop
+        )
+        await fut
+        yield fut.result()
+    else:
+        submitted = [
+            asyncio.wrap_future(
+                pool.submit(execute_one_plan, lmdb_environment, plan, log), loop=loop
             )
-
-        if len(tasks) == 1:
-            task = tasks[0]
-        else:
-            task = asyncio.gather(*tasks)
-        # log queries, analyze results, etc.
-        task.add_done_callback(analyze)
-        return task, events
-    except:
-        log.exception("executor")
+            for plan in plans
+        ]
+        for fut in submitted:
+            await fut
+            yield fut.result()
 
 
 def execute_one_plan(
-    lmdb_environment: lmdb.Environment, plan: QueryPlan, on_event, log
+    lmdb_environment: lmdb.Environment, plan: QueryPlan, log: logging.Logger
 ):
     """
     Run a single query plan, calling on_event for each event
     """
+    events = []
+    on_event = events.append
     try:
         limit = plan.limit
         count = 0
@@ -830,25 +1047,30 @@ def execute_one_plan(
         log.exception("execute_one_plan")
 
     finally:
-        return plan, log
+        return plan, events
 
 
-SLOW_QUERY_THRESHOLD = 500
+def analysis_thread(work_queue, slow_query_threshold=500, delay=0.5):
+    from time import sleep
+
+    log = logging.getLogger("nostr_relay.queries")
+    while True:
+        plans = work_queue.get()
+
+        sleep(delay)
+        try:
+            _analyze(plans, log, slow_query_threshold=slow_query_threshold)
+        except:
+            log.exception("analyze")
+        del plans
 
 
-def _analyze_threaded(task):
-    result = task.result()
-    if not isinstance(result, list):
-        plans, log = [result[0]], result[1]
-
-    else:
-        plans, log = [r[0] for r in result], result[0][1]
-
+def _analyze(plans, log, slow_query_threshold=500):
     for plan in plans:
         log.debug("Executed plan. Stats: %s", plan.stats)
         stats = plan.stats
         duration = (stats["end"] - stats["start"]) * 1000  # milliseconds
-        if duration > SLOW_QUERY_THRESHOLD:
+        if duration > slow_query_threshold:
             log.info(
                 "Slowish query: %s since:%s until:%s – took %.2fms",
                 plan.query,
@@ -869,14 +1091,28 @@ def _analyze_threaded(task):
                 )
 
 
-def analyze(task, later=True):
+ANALYSIS_QUEUE = SimpleQueue()
+
+
+def analyze(plans, later=True, log=None):
     """
     Analyze and log query stats from the completed task
     """
     if later:
-        QUERY_POOL.submit(_analyze_threaded, task)
+        if analyze.ANALYSIS_THREAD is None:
+            analyze.ANALYSIS_THREAD = threading.Thread(
+                target=analysis_thread,
+                args=(ANALYSIS_QUEUE,),
+                kwargs={"delay": Config.get("analysis_delay", 0.5)},
+            )
+            analyze.ANALYSIS_THREAD.daemon = True
+            analyze.ANALYSIS_THREAD.start()
+        ANALYSIS_QUEUE.put(plans)
     else:
-        _analyze_threaded(task)
+        _analyze(plans, log)
+
+
+analyze.ANALYSIS_THREAD = None
 
 
 def encode_event(event):
@@ -894,18 +1130,36 @@ def encode_event(event):
 
 
 def decode_event(data):
-    et = unpackb(data)
-    assert et[0] == 1, "unknown version"
-    event = Event(
-        id=et[1].hex(),
-        created_at=et[2],
-        kind=et[3],
-        pubkey=et[4].hex(),
-        content=et[5],
-        tags=et[6],
-        sig=et[7].hex(),
-    )
-    return event
+    if data:
+        assert data[0] == 1, "unknown version"
+        event = Event(
+            id=data[1].hex(),
+            created_at=data[2],
+            kind=data[3],
+            pubkey=data[4].hex(),
+            content=data[5],
+            tags=data[6],
+            sig=data[7].hex(),
+        )
+        return event
+
+
+def get_event_data(txn, event_id: bytes):
+    try:
+        return unpackb(txn.get(b"\x00" + event_id), use_list=False)
+    except TypeError:
+        return None
+
+
+def bytes_from_hex(hexstr: str):
+    try:
+        return bytes.fromhex(hexstr)
+    except ValueError:
+        if len(hexstr) >= 4 and len(hexstr) % 2 != 0:
+            hexstr = hexstr[:-1]
+            return bytes.fromhex(hexstr)
+        else:
+            raise
 
 
 if __name__ == "__main__":
