@@ -5,14 +5,20 @@ import falcon
 
 from time import time
 from falcon import media
-from websockets.exceptions import ConnectionClosedError, ConnectionClosedOK
+
+try:
+    from websockets.exceptions import ConnectionClosedError, ConnectionClosedOK
+except ImportError:
+    from wsproto.utilities import ProtocolError
+
+    ConnectionClosedError = ConnectionClosedOK = ProtocolError
 
 import falcon.asgi
 
 from .rate_limiter import get_rate_limiter
 from . import __version__
 from .config import Config
-from .util import timeout, json
+from .util import timeout, json_dumps, json_loads, event_as_json, JSONDecodeError
 from .errors import AuthenticationError, StorageError
 
 
@@ -22,28 +28,29 @@ class Client:
         "remote_addr",
         "log",
         "id",
-        "running",
         "subscription_queue",
         "send_task",
         "rate_limiter",
         "timeout",
         "auth_token",
         "sent",
+        "start_time",
     )
 
     def __init__(self, ws, req, rate_limiter=None, log=None, timeout=1800):
         self.ws = ws
-        self.remote_addr = req.remote_addr
-        self.id = f"{req.remote_addr}-{secrets.token_hex(2)}"
+        self.remote_addr = req.remote_addr if req else "%s:%s" % ws.remote_address
+        origin = req.get_header("origin") if req else ws.origin
+        self.id = f"{self.remote_addr}-{secrets.token_hex(2)}"
         self.log = log
-        self.log.info(f'Accepted {self.id} from Origin: {req.get_header("origin")}')
-        self.running = True
+        self.log.info(f"Accepted {self.id} from Origin: {origin}")
         self.subscription_queue = asyncio.Queue()
         self.send_task = None
         self.sent = 0
         self.rate_limiter = rate_limiter
         self.timeout = timeout
         self.auth_token = {}
+        self.start_time = time()
 
     def validate_message(self, message):
         if not isinstance(message, list):
@@ -54,35 +61,33 @@ class Client:
             return False
         return True
 
-    async def send_subscriptions(self):
-        ws = self.ws
+    async def send_subscriptions(self, ws_send):
         get_from_storage = self.subscription_queue.get
-        send_text = ws.send_text
-        sent = 0
         self.log.debug("%s waiting for subs", self.id)
-        while self.running and ws.ready:
+        while True:
             try:
                 sub_id, event = await get_from_storage()
                 if event is not None:
-                    message = event.to_message(sub_id)
+                    # message = json_dumps(["EVENT", sub_id, event.to_json_object()])
+                    message = event_as_json(sub_id, event)
                 else:
                     # done with stored events
-                    message = f'["EOSE", "{sub_id}"]'
-                await send_text(message)
+                    message = f'["EOSE","{sub_id}"]'
+
+                await ws_send(message)
                 # self.log.debug("SENT: %s", message)
                 self.sent += len(message)
             except (
                 falcon.WebSocketDisconnected,
                 ConnectionClosedError,
                 ConnectionClosedOK,
+                asyncio.CancelledError,
             ):
                 break
-            except asyncio.CancelledError:
-                break
             except Exception:
                 self.log.exception("subs")
 
-    async def start(self, storage):
+    async def start(self, storage, ws_send, ws_recv):
         ws = self.ws
         client_id = self.id
         remote_addr = self.remote_addr
@@ -90,15 +95,15 @@ class Client:
         if storage.authenticator.is_enabled:
             challenge = storage.authenticator.get_challenge(remote_addr)
             self.log.debug("Sent challenge %s to %s", challenge, client_id)
-            await ws.send_media(["AUTH", challenge])
+            await ws_send(json_dumps(["AUTH", challenge]))
             throttle = await storage.authenticator.should_throttle(self.auth_token)
         else:
             throttle = 0
 
-        while ws.ready:
+        while True:
             try:
                 async with timeout(self.timeout):
-                    message = await ws.receive_media()
+                    message = json_loads(await ws_recv())
 
                 if not self.validate_message(message):
                     if throttle:
@@ -120,17 +125,11 @@ class Client:
                         ]
                     else:
                         response = ["NOTICE", "rate-limited"]
-                    await ws.send_media(response)
+                    await ws_send(json_dumps(response))
                     continue
 
                 if command == "REQ":
-                    if throttle:
-                        await asyncio.sleep(throttle)
-
                     sub_id = str(message[1])
-                    if self.send_task is None:
-                        self.send_task = asyncio.create_task(self.send_subscriptions())
-                        await asyncio.sleep(0)
                     await storage.subscribe(
                         client_id,
                         sub_id,
@@ -138,19 +137,22 @@ class Client:
                         self.subscription_queue,
                         auth_token=self.auth_token,
                     )
+                    if throttle:
+                        await asyncio.sleep(throttle)
+                    if self.send_task is None:
+                        self.send_task = asyncio.create_task(
+                            self.send_subscriptions(ws_send)
+                        )
                 elif command == "CLOSE":
                     sub_id = str(message[1])
                     await storage.unsubscribe(client_id, sub_id)
                 elif command == "EVENT":
-                    if throttle:
-                        await asyncio.sleep(throttle)
-
                     try:
                         event, result = await storage.add_event(
                             message[1], auth_token=self.auth_token
                         )
                     except (StorageError, AuthenticationError) as e:
-                        throttle = min(throttle, 1) * 2
+                        throttle = max(throttle, 1) * 2
                         self.log.error("Auth error %s. Throttling %s", str(e), throttle)
                         result = False
                         reason = str(e)
@@ -167,7 +169,9 @@ class Client:
                             "%s added %s from %s", client_id, event.id, event.pubkey
                         )
                     finally:
-                        await ws.send_media(["OK", eventid, result, reason])
+                        if throttle:
+                            await asyncio.sleep(throttle)
+                        await ws_send(json_dumps(["OK", eventid, result, reason]))
                 elif command == "AUTH" and storage.authenticator.is_enabled:
                     self.auth_token = await storage.authenticator.authenticate(
                         message[1], challenge=challenge
@@ -175,30 +179,47 @@ class Client:
                     throttle = await storage.authenticator.should_throttle(
                         self.auth_token
                     )
+
             except StorageError as e:
                 self.log.warning("storage error: %s for %s", e, client_id)
-                await ws.send_media(["NOTICE", str(e)])
+                await ws_send(json_dumps(["NOTICE", str(e)]))
             except AuthenticationError as e:
                 self.log.warning("Auth error. %s token:%s", str(e), self.auth_token)
-                await ws.send_media(["NOTICE", str(e)])
-            except (falcon.WebSocketDisconnected, ConnectionClosedError):
+                await ws_send(json_dumps(["NOTICE", str(e)]))
+            except (
+                falcon.WebSocketDisconnected,
+                ConnectionClosedError,
+                ConnectionClosedOK,
+            ):
                 break
-            except json.JSONDecodeError:
+            except JSONDecodeError:
                 self.log.debug("json decoding")
                 continue
             except asyncio.TimeoutError:
                 self.log.info("%s timed out", client_id)
                 await ws.close(code=1013)
                 break
+            except Exception:
+                self.log.exception("client loop")
+                await ws.close(code=1013)
+                break
+        await storage.unsubscribe(self.id)
 
     async def stop(self):
-        self.running = False
         if self.send_task:
             self.send_task.cancel()
             try:
                 await self.send_task
             except asyncio.CancelledError:
                 pass
+        self.rate_limiter.cleanup()
+        duration = time() - self.start_time
+        self.log.info(
+            "Done %s. Sent: %d Bytes. Duration: %d Seconds",
+            self,
+            self.sent,
+            duration,
+        )
 
     def __str__(self):
         return f"{self.auth_token.get('pubkey', 'anon')}@{self.id}"
@@ -224,6 +245,8 @@ class NostrAPI(BaseResource):
             supported_nips = [1, 2, 5, 9, 11, 12, 15, 20, 26, 33, 40]
             if Config.authentication.get("enabled"):
                 supported_nips.append(42)
+            if Config.fts_enabled:
+                supported_nips.append(50)
             metadata = {
                 "name": Config.relay_name,
                 "description": Config.relay_description,
@@ -258,7 +281,6 @@ class NostrAPI(BaseResource):
                 self.log.warning("rate-limited ACCEPT %s", req.remote_addr)
                 return
             await ws.accept()
-            start = time()
 
         except falcon.WebSocketDisconnected:
             return
@@ -275,27 +297,15 @@ class NostrAPI(BaseResource):
             # from .util import easy_profiler
 
             # with easy_profiler():
-            await client.start(self.storage)
+            await client.start(self.storage, ws.send_text, ws.receive_text)
         except (
             falcon.WebSocketDisconnected,
             ConnectionClosedError,
             ConnectionClosedOK,
         ):
             pass
-        except Exception:
-            await ws.close(code=1013)
-            self.log.exception("client loop")
         finally:
             await client.stop()
-            await self.storage.unsubscribe(client.id)
-            self.rate_limiter.cleanup()
-            duration = time() - start
-            self.log.info(
-                "Done %s. Sent: %d Bytes. Duration: %d Seconds",
-                client,
-                client.sent,
-                duration,
-            )
 
 
 class NostrStats(BaseResource):
@@ -378,7 +388,6 @@ def create_app(conf_file=None, storage=None):
     import os
     import os.path
     import logging, logging.config
-    from functools import partial
 
     Config.load(conf_file)
     if Config.DEBUG:
@@ -399,13 +408,14 @@ def create_app(conf_file=None, storage=None):
     rate_limiter = get_rate_limiter(Config)
 
     json_handler = media.JSONHandlerWS(
-        dumps=partial(json.dumps, ensure_ascii=False),
-        loads=json.loads,
+        dumps=json_dumps,
+        loads=json_loads,
     )
 
     logging.info("Starting version %s", __version__)
+    api = NostrAPI(store, rate_limiter=rate_limiter)
     app = falcon.asgi.App(middleware=SetupMiddleware(store))
-    app.add_route("/", NostrAPI(store, rate_limiter=rate_limiter))
+    app.add_route("/", api)
     app.add_route("/stats/", NostrStats(store))
     app.add_route("/e/{event_id}", ViewEventResource(store))
     app.add_route("/.well-known/nostr.json", NostrIDP(store))
@@ -426,6 +436,9 @@ def run_with_gunicorn(conf_file=None):
             import sys
 
             if sys.implementation.name == "pypy":
+                from uvicorn.workers import UvicornH11Worker
+
+                # UvicornH11Worker.CONFIG_KWARGS["ws"] = "wsproto"
                 worker_class = "uvicorn.workers.UvicornH11Worker"
             else:
                 worker_class = "uvicorn.workers.UvicornWorker"
