@@ -13,6 +13,7 @@ from nostr_relay.config import Config
 from nostr_relay.web import create_app
 from nostr_relay.storage import get_storage
 from nostr_relay.auth import Authenticator
+from nostr_relay.util import timeout
 from aionostr.event import Event, PrivateKey
 from nostr_relay.errors import StorageError
 
@@ -518,11 +519,18 @@ class MainTests(APITests):
             for event in EVENTS:
                 await self.send_event(ws, event, True)
             await ws.send_json(
-                ["REQ", "test", {"ids": [EVENTS[1]["id"]]}, {"#p": EVENTS[1]["pubkey"]}]
+                [
+                    "REQ",
+                    "test",
+                    {"ids": [EVENTS[1]["id"]]},
+                    {"#p": [EVENTS[1]["pubkey"]]},
+                ]
             )
             data = await ws.receive_json()
             assert data == ["EVENT", "test", EVENTS[1]]
             data = await ws.receive_json()
+            assert data == ["EVENT", "test", EVENTS[2]]
+            data = await ws.receive_json()
             assert data == ["EOSE", "test"]
             await ws.send_json(["CLOSE", "test"])
 
@@ -567,22 +575,23 @@ class MainTests(APITests):
                 ]
             )
 
-            data = await ws.receive_json()
-            assert data == ["EVENT", "test", EVENTS[1]]
+            async with timeout(5):
+                data = await ws.receive_json()
+                assert data == ["EVENT", "test", EVENTS[1]]
 
-            data = await ws.receive_json()
-            assert data == ["EOSE", "test"]
+                data = await ws.receive_json()
+                assert data == ["EOSE", "test"]
 
-            await ws.send_json(["REQ", "missing", {"kinds": [1000]}])
-            data = await ws.receive_json()
-            assert data == ["EOSE", "missing"]
+                await ws.send_json(["REQ", "missing", {"kinds": [1000]}])
+                data = await ws.receive_json()
+                assert data == ["EOSE", "missing"]
 
-            # now add a new event
-            data = await self.send_event(ws, EVENTS[2], True)
-            assert data[0] == "OK"
-            await asyncio.sleep(1)
-            data = await ws.receive_json()
-            assert data == ["EVENT", "test", EVENTS[2]]
+                # now add a new event
+                data = await self.send_event(ws, EVENTS[2], True)
+                assert data[0] == "OK"
+                await asyncio.sleep(1)
+                data = await ws.receive_json()
+                assert data == ["EVENT", "test", EVENTS[2]]
 
     async def test_req_after_add_bad(self):
         """
@@ -608,7 +617,7 @@ class MainTests(APITests):
             assert data[0] == "OK"
             await asyncio.sleep(0.5)
             with self.assertRaises(asyncio.TimeoutError):
-                async with asyncio.timeout(1.2):
+                async with timeout(1.2):
                     data = await ws.receive_json()
 
     async def test_prefix_search(self):
@@ -936,7 +945,7 @@ class AuthTests(APITests):
 
             # unauthenticated throttled 1 second
             with self.assertRaises(asyncio.TimeoutError):
-                async with asyncio.timeout(0.5):
+                async with timeout(0.5):
                     response = await self.send_event(ws, EVENTS[2], True)
             await ws.receive_json()
 
@@ -950,7 +959,7 @@ class AuthTests(APITests):
                 ]
             )
             with self.assertRaises(asyncio.TimeoutError):
-                async with asyncio.timeout(2.0):
+                async with timeout(2.0):
                     response = await self.send_event(ws, EVENTS[2], True)
             await ws.receive_json()
 
@@ -963,7 +972,7 @@ class AuthTests(APITests):
                     ),
                 ]
             )
-            async with asyncio.timeout(0.5):
+            async with timeout(0.5):
                 response = await self.send_event(ws, EVENTS[2], True)
 
 
